@@ -6,6 +6,7 @@ import Test.Framework.Providers.HUnit
 
 import Test.HUnit
 
+import Control.Concurrent.STM
 import Control.Concurrent.Async
 import Control.Exception
 import Data.IORef
@@ -29,6 +30,7 @@ tests = [
   , testGroup "async_cancel_rep" $
       replicate 1000 $
          testCase "async_cancel"       async_cancel
+  , testCase "async_cancelmany" async_cancelmany
   , testCase "async_poll"        async_poll
   , testCase "async_poll2"       async_poll2
   , testCase "withasync_waitCatch_blocked" withasync_waitCatch_blocked
@@ -43,6 +45,10 @@ tests = [
   , testCase "concurrently_" case_concurrently_
   , testCase "replicateConcurrently_" case_replicateConcurrently
   , testCase "replicateConcurrently" case_replicateConcurrently_
+  , testCase "link1" case_link1
+  , testCase "link2" case_link2
+  , testCase "link1_cancel" case_link1cancel
+  , testCase "concurrently_deadlock" case_concurrently_deadlock
  ]
 
 value = 42 :: Int
@@ -90,7 +96,7 @@ withasync_wait2 = do
   a <- withAsync (threadDelay 1000000) $ return
   r <- waitCatch a
   case r of
-    Left e  -> fromException e @?= Just ThreadKilled
+    Left e  -> fromException e @?= Just AsyncCancelled
     Right _ -> assertFailure ""
 
 async_cancel :: Assertion
@@ -102,6 +108,18 @@ async_cancel = do
     Left e -> fromException e @?= Just TestException
     Right r -> r @?= value
 
+async_cancelmany :: Assertion -- issue 59
+async_cancelmany = do
+  r <- newIORef []
+  a <- async $ forConcurrently_ ['a'..'z'] $ \c ->
+    delay 2 `finally` atomicModifyIORef r (\i -> (c:i,()))
+  delay 1
+  cancel a
+  v <- readIORef r
+  assertEqual "cancelmany" 26 (length v)
+  where
+    delay sec = threadDelay (sec * 1000000)
+
 async_poll :: Assertion
 async_poll = do
   a <- async (threadDelay 1000000)
@@ -236,3 +254,86 @@ case_replicateConcurrently_ = do
   () <- replicateConcurrently_ 100 action
   resVal <- readIORef ref
   resVal @?= 100
+
+case_link1 :: Assertion
+case_link1 = do
+  m1 <- newEmptyMVar
+  m2 <- newEmptyMVar
+  let ex = ErrorCall "oops"
+  a <- async $ do takeMVar m1; throwIO ex; putMVar m2 ()
+  link a
+  e <- try $ (do
+    putMVar m1 ()
+    takeMVar m2)
+  assertBool "link1" $
+    case e of
+      Left (ExceptionInLinkedThread a' e') ->
+        compareAsyncs a' a == EQ &&
+          case fromException e' of
+            Just (ErrorCall s) -> s == "oops"
+            _otherwise -> False
+      _other -> False
+
+case_link2 :: Assertion
+case_link2 = do
+  let
+    setup = do
+      m1 <- newEmptyMVar
+      m2 <- newEmptyMVar
+      let ex1 = ErrorCall "oops1"; ex2 = ErrorCall "oops2"
+      a <- async $ do takeMVar m1; throwIO ex1
+      b <- async $ do takeMVar m2; throwIO ex2
+      link2 a b
+      return (m1,m2,a,b)
+
+  (m1,m2,a,b) <- setup
+  e <- try $ do
+    putMVar m1 ()
+    wait b
+  putMVar m2 ()  -- ensure the other thread is not deadlocked
+  assertBool "link2a" $
+    case e of
+      Left (ExceptionInLinkedThread a' e') ->
+        compareAsyncs a' a == EQ &&
+          case fromException e' of
+            Just (ErrorCall s) -> s == "oops1"
+            _otherwise -> False
+      _other -> False
+
+  (m1,m2,a,b) <- setup
+  e <- try $ do
+    putMVar m2 ()
+    wait a
+  putMVar m1 ()  -- ensure the other thread is not deadlocked
+  assertBool "link2b" $
+    case e of
+      Left (ExceptionInLinkedThread a' e') ->
+        compareAsyncs a' b == EQ &&
+          case fromException e' of
+            Just (ErrorCall s) -> s == "oops2"
+            _otherwise -> False
+      _other -> False
+
+case_link1cancel :: Assertion
+case_link1cancel = do
+  m1 <- newEmptyMVar
+  let ex = ErrorCall "oops"
+  a <- async $ do takeMVar m1
+  link a
+  e <- try $ do cancel a; wait a
+  putMVar m1 ()
+  assertBool "link1cancel" $
+    case e of
+      Left AsyncCancelled -> True  -- should not be ExceptionInLinkedThread
+      _other -> False
+
+-- See Issue #62
+case_concurrently_deadlock :: Assertion
+case_concurrently_deadlock = do
+  tvar <- newTVarIO False :: IO (TVar Bool)
+  e <- try $ void $ join (concurrently) (atomically $ readTVar tvar >>= check)
+    -- should throw BlockedIndefinitelyOnSTM not BlockedIndefinitelyOnMVar
+  assertBool "concurrently_deadlock" $
+    case e of
+      Left BlockedIndefinitelyOnSTM{} -> True
+      _other -> False
